import random
from typing import Sequence


class HSLA:
    """
    `HSLA` (Hue, Saturation, Lightness, Alpha) color model.
    """
    __slots__ = ('_h', '_s', '_l', '_a')

    def __init__(self, color: Sequence, /) -> None:
        """
        `RGBA` color constructor.

        Attributes
        ----------
        color: `Sequence`
            Color sequence of h, s, l and optional a.

        Raises
        ------
        `ValueError` if the color is invalid.
        """
        match color:
            case tuple() | list():
                self.h = color[0]
                self.s = color[1]
                self.l = color[2]
                self.a = color[3] if len(color) == 4 else 100
            case _:
                raise ValueError(f"invalid color value: {color}")
    
    # magic methods
    def __eq__(self, other) -> bool:
        return isinstance(other, HSLA) and self.hsla == other.hsla

    def __ne__(self, other) -> bool:
        return not self.__eq__(other)

    def __str__(self) -> str:
        return f"hsla{self.hsla}"

    def __repr__(self) -> str:
        return f"<HSLA h={self.h}, s={self.s}, l={self.l}, a={self.a}>"

    def __hash__(self) -> int:
        h = hash(self.h)
        s = hash(self.s)
        l = hash(self.l)
        a = hash(self.a)

        return h ^ s ^ l ^ a
            
    def __getitem__(self, key):
        return self.hsla[key]
    
    def __iter__(self):
        for item in self.hsla:
            yield item

    # attributes
    @property
    def h(self) -> int:
        """
        Hue value in range `0-359`.
        """
        return self._h
    
    @h.setter
    def h(self, value: int):
        self._h = round(value % 360)

    hue = h

    @property
    def s(self) -> int:
        """
        Saturation value in range `0-100`.
        """
        return self._s
    
    @s.setter
    def s(self, value: int):
        self._s = min(round(value), 100)

    saturation = s

    @property
    def l(self) -> int:
        """
        Lightness value in range `0-100`.
        """
        return self._l
    
    @l.setter
    def l(self, value: int):
        self._l = min(round(value), 100)

    lightness = l

    @property
    def a(self) -> int:
        """
        Alpha value (transparency) in range `0-100`.
        """
        return self._a
    
    @a.setter
    def a(self, value: int):
        self._a = min(round(value), 100)

    alpha = a

    # formats
    @property
    def hsl(self) -> tuple[int, int, int]:
        """
        Color as `(h, s, l)` tuple.
        """
        return (self.h, self.s, self.l)
    
    @property
    def hsla(self) -> tuple[int, int, int, int]:
        """
        Color as `(h, s, l, a)` tuple.
        """
        return (self.h, self.s, self.l, self.a)
    
    # converters
    def copy(self) -> "HSLA":
        """
        Get a copy of the color.
        """
        obj = HSLA.__new__(HSLA)
        obj._h = self._h
        obj._s = self._s
        obj._l = self._l
        obj._a = self._a

        return obj

    def to_rgba(self):
        """
        Convert the color to `RGBA` model.
        """
        from .rgba import RGBA

        h = self.h / 360.0
        s = self.s / 100.0
        l = self.l / 100.0
        
        def hue_to_rgb(p, q, t):
            if t < 0:
                t += 1
            if t > 1:
                t -= 1
            if t < 1/6:
                return p + (q - p) * 6 * t
            if t < 1/2:
                return q
            if t < 2/3:
                return p + (q - p) * (2/3 - t) * 6
            return p

        if s == 0:
            r = g = b = int(l * 255)
        else:
            q = l * (1 + s) if l < 0.5 else l + s - l * s
            p = 2 * l - q
            r = hue_to_rgb(p, q, h + 1/3) * 255
            g = hue_to_rgb(p, q, h) * 255
            b = hue_to_rgb(p, q, h - 1/3) * 255

        return RGBA((r, g, b, self.a * 2.55))

    # utils
    def range(self, num: int, step: int, angle: int) -> list["HSLA"]:
        """
        Get a list of circular colors.

        Attributes
        ----------
        num: `int`
            Number of colors.
        step: `int`
            Angle in degrees.
        angle: `int`
            Start angle.
        """
        result = []

        for i in range(num):
            co = self.copy()
            co.h = angle + step * i
            result.append(co)

        return result

    def complementary(self) -> "HSLA":
        """
        Get a complementary color.
        """
        color = self.copy()
        color.h += 180
        return color
    
    def split_complementary(self) -> list["HSLA"]:
        """
        Get 2 split complementary colors.
        """
        return self.range(2, 60, self.h + 150)
    
    def triadic(self) -> list["HSLA"]:
        """
        Get 2 triadic colors.
        """
        return self.range(2, 120, self.h + 120)
    
    def tetradic(self) -> list["HSLA"]:
        """
        Get 3 tetradic colors.
        """
        return self.range(3, 90, self.h + 90)
    
    def analogous(self) -> list["HSLA"]:
        """
        Get 3 analogous colors.
        """
        return self.range(3, 30, self.h - 30)
    
    # color generators
    @staticmethod
    def random() -> "HSLA":
        return HSLA([
            random.randint(0, i)
            for i in (360, 100, 100, 100)
        ])


import random
from typing import Sequence


class CMYK:
    """
    `CMYK` (Cyan, Magenta, Yellow, Black Key) color model.
    """
    __slots__ = ('_c', '_m', '_y', '_k')

    def __init__(self, color: Sequence, /) -> None:
        """
        `CMYK` color constructor.

        Attributes
        ----------
        color: `Sequence`
            Color sequence of c, m, y, k.

        Raises
        ------
        `ValueError` if the color is invalid.
        """
        match color:
            case tuple() | list():
                self.c = color[0]
                self.m = color[1]
                self.y = color[2]
                self.k = color[3]
            case _:
                raise ValueError(f"invalid color value: {color}")
            
    # magic methods
    def __eq__(self, other) -> bool:
        return isinstance(other, CMYK) and self.cmyk == other.cmyk

    def __ne__(self, other) -> bool:
        return not self.__eq__(other)

    def __str__(self) -> str:
        return f"cmyk{self.cmyk}"

    def __repr__(self) -> str:
        return f"<CMYK c={self.c}, m={self.m}, y={self.y}, k={self.k}>"

    def __hash__(self) -> int:
        c = hash(self.c)
        m = hash(self.m)
        y = hash(self.y)
        k = hash(self.k)

        return c ^ m ^ y ^ k
            
    def __getitem__(self, key):
        return self.cmyk[key]
    
    def __iter__(self):
        for item in self.cmyk:
            yield item
            
    # attributes
    @property
    def c(self) -> int:
        """
        Cyan value in range `0-100`.
        """
        return self._c
    
    @c.setter
    def c(self, value: int):
        self._c = min(round(value), 100)

    cyan = c
            
    @property
    def m(self) -> int:
        """
        Magenta value in range `0-100`.
        """
        return self._m
    
    @m.setter
    def m(self, value: int):
        self._m = min(round(value), 100)

    magenta = m

    @property
    def y(self) -> int:
        """
        Yellow value in range `0-100`.
        """
        return self._y
    
    @y.setter
    def y(self, value: int):
        self._y = min(round(value), 100)

    yellow = y

    @property
    def k(self) -> int:
        """
        Black key in range `0-100`.
        """
        return self._k
    
    @k.setter
    def k(self, value: int):
        self._k = min(round(value), 100)

    key = k
    black = k

    @property
    def cmyk(self) -> tuple[int, int, int, int]:
        """
        Color as `(c, m, y, k)` tuple.
        """
        return (self.c, self.m, self.y, self.k)
    
    # converters
    def copy(self) -> "CMYK":
        """
        Get a copy of the color.
        """
        obj = CMYK.__new__(CMYK)
        obj._c = self._c
        obj._m = self._m
        obj._y = self._y
        obj._k = self._k

        return obj

    def to_rgba(self):
        """
        Convert the color to `RGBA` model.
        """
        from .rgba import RGBA

        return RGBA([
            255 * (1 - i / 100) * (1 - self.k / 100)
            for i in self.cmyk[:3]
        ])
    
    # color generators
    @staticmethod
    def random() -> "CMYK":
        return CMYK([
            random.randint(0, 100)
            for _ in range(4)
        ])
    


import math
from typing import Iterable


def distance(first: Iterable, second: Iterable):
    """
    Get Euclidean distance for 2 iterables.
    """
    return math.sqrt(sum((a - b) ** 2 for a, b in zip(first, second)))


import random
from typing import Sequence

from .utils import distance


class RGBA:
    """
    `RGBA` (Red, Green, Blue, Alpha) color model.
    """
    __slots__ = ('value', '_bits', '_max_one', '_max_all')

    def __init__(self, color: int | str | Sequence, /, bits: int = 8) -> None:
        """
        `RGBA` color constructor.

        Attributes
        ----------
        color: `int` | `str` | `Sequence`
            Color value, hex or sequence of r, g, b and optional a.
        bits: `int`
            Number of bits per channel. Must be factor of 4.
            Default is 8 bits that equals 256 values per channel.

        Raises
        ------
        `ValueError` if the color is invalid.
        """
        self.bits = bits
        self._max_one = 2 ** bits - 1
        self._max_all = 2 ** (bits * 4) - 1

        match color:
            case str():
                num = len(color)
                chars = bits // 4
                if num not in {bits, chars * 3}:
                    raise ValueError(f"invalid hex value: {color}")
                
                self.value = (
                    (int(color[:chars], base=16))
                    + (int(color[chars : chars * 2], base=16) << bits)
                    + (int(color[chars * 2 : chars * 3], base=16) << (bits * 2))
                    + ((int(color[chars * 3 :], base=16) if num == bits else self._max_one) << bits * 3)
                )

            case int():
                self.value = color
        
            case list() | tuple():
                num = len(color)
                if num not in {3, 4}:
                    raise ValueError(f"invalid color sequence: {color}")

                self.value = (
                    min(int(color[0]), self._max_one)
                    + (min(int(color[1]), self._max_one) << bits) 
                    + (min(int(color[2]), self._max_one) << bits * 2) 
                    + ((min(int(color[3]), self._max_one) if num == 4 else self._max_one) << bits * 3)
                )

            case _:
                raise ValueError(f"invalid color value: {color}")
    
    # magic methods
    def __eq__(self, other) -> bool:
        return isinstance(other, RGBA) and self.value == other.value

    def __ne__(self, other) -> bool:
        return not self.__eq__(other)

    def __str__(self) -> str:
        return f"rgba{self.rgba}"

    def __int__(self) -> int:
        return self.value

    def __repr__(self) -> str:
        return f"<RGBA value={self.value}, bits={self._bits}>"

    def __hash__(self) -> int:
        return hash(self.value)
    
    def __getitem__(self, key):
        return self.rgba[key]
    
    def __iter__(self):
        for item in self.rgba:
            yield item
    
    # private methods
    def _bit_mask(self, pos):
        n = self._bits
        res = bin(self._max_all)
        res = res[:n * pos + 2] + '0' * n + res[n * (pos + 1) + 2:]
        return int(res, 2)

    # attributes
    @property
    def bits(self):
        return self._bits
    
    @bits.setter
    def bits(self, value: int):
        if value % 4 != 0 or value < 4:
            raise ValueError(f"number of bits must be factor of 4")
        
        self._bits = value

    @property
    def r(self) -> int:
        """
        Red value.
        """
        return (self.value) & self._max_one

    @r.setter
    def r(self, value: int):
        self.value = (self.value & self._bit_mask(3)) | min(int(value), self._max_one)

    red = r

    @property
    def g(self) -> int:
        """
        Green value.
        """
        return (self.value >> self._bits) & self._max_one
    
    green = g

    @g.setter
    def g(self, value: int):
        self.value = (self.value & self._bit_mask(2)) | (min(int(value), self._max_one) << self._bits)

    @property
    def b(self) -> int:
        """
        Blue value.
        """
        return (self.value >> self._bits * 2) & self._max_one

    @b.setter
    def b(self, value: int):
        self.value = (self.value & self._bit_mask(1)) | (min(int(value), self._max_one) << self._bits * 2)

    blue = b

    @property
    def a(self) -> int:
        """
        Alpha value (transparency).
        """
        return (self.value >> self._bits * 3) & self._max_one

    @a.setter
    def a(self, value: int):
        self.value = (self.value & self._bit_mask(0)) | (min(int(value), self._max_one) << self._bits * 3)

    alpha = a

    # formats
    @property
    def rgb(self) -> tuple[int, int, int]:
        """
        Color as `(r, g, b)` tuple.
        """
        return (self.r, self.g, self.b)

    @property
    def rgba(self) -> tuple[int, int, int, int]:
        """
        Color as `(r, g, b, a)` tuple.
        """
        return (self.r, self.g, self.b, self.a)

    @property
    def hex(self) -> str:
        """
        Color as `'rrggbb'` string.
        """
        return f'{self.r:02X}{self.g:02X}{self.b:02X}'
    
    @property
    def hexa(self) -> str:
        """
        Color as 'rrggbbaa' string.
        """
        return f'{self.r:02X}{self.g:02X}{self.b:02X}{self.a:02X}'
    
    @property
    def brightness(self) -> int:
        """
        Get a perceived brightness of the color.
        `HSP` color model is used for the calculation.
        """
        return round(
            (
                0.299 * ((self.r / self._max_one) ** 2) 
                + 0.587 * ((self.g / self._max_one) ** 2) 
                + 0.114 * ((self.b / self._max_one) ** 2)
            ) 
            * self._max_one
        )

    # converters
    def copy(self):
        """
        Get a copy of the color.
        """
        obj = RGBA.__new__(RGBA)
        obj.value = self.value
        return obj
    
    def to_hsla(self):
        """
        Convert the color to `HSLA` model.
        """
        from .hsla import HSLA

        r = self.r / self._max_one
        g = self.g / self._max_one
        b = self.b / self._max_one

        cmax = max(r, g, b)
        cmin = min(r, g, b)
        delta = cmax - cmin

        l = (cmax + cmin) / 2.0

        s = 0.0
        if delta != 0.0:
            s = delta / (1 - abs(2 * l - 1))

        h = 0.0
        if delta != 0.0:
            if cmax == r:
                h = 60 * ((g - b) / delta % 6)
            elif cmax == g:
                h = 60 * ((b - r) / delta + 2)
            elif cmax == b:
                h = 60 * ((r - g) / delta + 4)

        return HSLA((h, s * 100, l * 100, self.a / self._max_one * 100))
    
    def to_cmyk(self):
        """
        Convert the color to `CMYK` model.
        """
        from .cmyk import CMYK

        max_channel = max(self.rgb)
        k = 1 - max_channel / self._max_one

        if k == 1:
            return CMYK((0, 0, 0, 100))

        return CMYK([
            (1 - i / self._max_one - k) / (1 - k) * 100
            for i in self.rgb
        ] + [k * 100])
    
    def to_web(self):
        """
        Get the closest web-safe color.
        """
        from .palette import Palette

        return self.closest(Palette.web()._items)
    
    def convert(self, bits: int):
        """
        Convert the color to another bit count.

        Attributes
        ----------
        bits: `int`
            Number of bits per channel. Must be factor of 4. 
        """
        scale = 2 ** (bits - self.bits)
        maxv = 2 ** bits - 1

        return RGBA([
            min(i * scale, maxv)
            for i in self.rgba
        ], bits=bits)

    # utils
    def is_light(self, threshold: int | None = None) -> bool:
        """
        Determines if color is light based on HSP color model.

        Attributes
        ----------
        threshold: `int` | `None`
            If brightness is greater than it, method will return `True`.
            If `None`, equals to half of the max value.
        """
        return self.brightness > (self._max_one / 2 if threshold is None else threshold) ** 2
    
    def is_web(self) -> bool:
        """
        Determines if color is web-safe.
        """
        from .palette import Palette

        return self in Palette.web()
            
    def complementary(self) -> "RGBA":
        """
        Get a complementary color.
        """
        return self.to_hsla().complementary().to_rgba()
    
    def split_complementary(self) -> list["RGBA"]:
        """
        Get 2 split complementary colors.
        """
        return [i.to_rgba() for i in self.to_hsla().split_complementary()]
    
    def triadic(self) -> list["RGBA"]:
        """
        Get 2 triadic colors.
        """
        return [i.to_rgba() for i in self.to_hsla().triadic()]
    
    def tetradic(self) -> list["RGBA"]:
        """
        Get 3 tetradic colors.
        """
        return [i.to_rgba() for i in self.to_hsla().tetradic()]
    
    def analogous(self) -> list["RGBA"]:
        """
        Get 3 analogous colors.
        """
        return [i.to_rgba() for i in self.to_hsla().analogous()]
    
    def closest(self, colors: list["RGBA"]) -> "RGBA":
        """
        Select the closest color to this from the list.

        Attributes
        ----------
        colors: list[RGBA]
            List of colors.

        Raises
        ------
        `ValueError` if the list of colors is empty.
        """        
        if len(colors) == 0:
            raise ValueError("unable to choose a color from an empty list")

        return min(colors, key=lambda c: distance(self, c))
    
    def futhest(self, colors: list["RGBA"]) -> "RGBA":
        """
        Select the futhest color to this from the list.

        Attributes
        ----------
        colors: list[RGBA]
            List of colors.

        Raises
        ------
        `ValueError` if the list of colors is empty.
        """        
        if len(colors) == 0:
            raise ValueError("unable to choose a color from an empty list")

        return max(colors, key=lambda c: distance(self, c))
    
    def blend(self, other: "RGBA", mode) -> "RGBA":
        """
        Blend the color with another color. 
        Note that the other color will be placed on top of this.

        Attributes
        ----------
        other: `RGBA`
            Foreground color.
        mode: `BlendMode`
            Blending mode. Must be a child class of `BlendMode`.

        Raises
        ------
        `ValueError` if the mode is invalid or bit counts of the colors do not match.
        """
        from .blend import BlendMode

        if not issubclass(mode, BlendMode):
            raise ValueError(f"blend mode must be inherited from the {BlendMode.__name__} class")

        return mode(self.bits).compose(self, other)
    
    # color generators
    @staticmethod
    def random(bits: int = 8) -> "RGBA":
        return RGBA([
            random.randint(0, 2 ** bits - 1)
            for _ in range(4)
        ], bits=bits)


Color = RGBA


class BlendMode:
    """
    The base class for blending modes. 
    Inherited classes must have overwritten `blend()` method.
    """
    def __init__(self, bits: int) -> None:
        self.bits = bits
        self.max_one = 2 ** bits - 1
        self.max_all = 2 ** (bits * 4) - 1

    def blend(self, bg: float, fg: float, bg_a: float, fg_a: float) -> float:
        """
        Calculate the result value for each channel.
        All arguments are in range `0-1`.

        Attributes
        ----------
        bg: `float`
            Background channel value.
        fg: `float`
            Foreground channel value.
        bg_a: `float`
            Background alpha value.
        fg_a: `float`
            Foreground alpha value.
        """
        raise NotImplementedError("blend mode class must have implemented 'blend' method")
    
    def comp(self, co: float, a: float) -> float:
        """
        Get a premultiplied color value with its complementary alpha.

        Attributes
        ----------
        co: `float`
            Color channel value.
        a: `float`
            Alpha value.
        """
        return co * (1 - a)
    
    def compose(self, bg, fg):
        """
        Compose 2 colors.

        Attributes
        ----------
        bg: `RGBA`
            Background color.
        fg: `RGBA`
            Foreground color.

        Raises
        ------
        `ValueError` if bit counts of the colors do not match.
        """
        from .rgba import RGBA

        if bg.bits != fg.bits:
            raise ValueError(f"can not blend colors with different size")
        
        maxv = self.max_one
        bg_a = bg.a / maxv
        fg_a = fg.a / maxv

        return RGBA((
            self.blend(bg.r / maxv, fg.r / maxv, bg_a, fg_a) * maxv,
            self.blend(bg.g / maxv, fg.g / maxv, bg_a, fg_a) * maxv,
            self.blend(bg.b / maxv, fg.b / maxv, bg_a, fg_a) * maxv,
            (bg_a + fg_a - bg_a * fg_a) * maxv
        ), self.bits)
        

class Normal(BlendMode):
    """
    Normal blending mode.
    """
    def blend(self, bg: float, fg: float, bg_a: float, fg_a: float) -> float:
        return fg * fg_a + self.comp(bg, fg_a)
    

class Darken(BlendMode):
    """
    Darken blending mode.
    """
    def blend(self, bg: float, fg: float, bg_a: float, fg_a: float) -> float:
        return min(fg * bg_a, bg * fg_a) + self.comp(fg, bg_a) + self.comp(bg, fg_a);


class Multiply(BlendMode):
    """
    Multiply blending mode.
    """
    def blend(self, bg: float, fg: float, bg_a: float, fg_a: float) -> float:
        return fg * bg + self.comp(fg, bg_a) + self.comp(bg, fg_a)
    

class ColorBurn(BlendMode):
    """
    Color Burn blending mode.
    """
    def blend(self, bg: float, fg: float, bg_a: float, fg_a: float) -> float:
        if fg == 0:
            if bg == bg_a:
                return fg_a * bg_a + self.comp(bg, fg_a)
            else:
                return self.comp(bg, fg_a)
        else:
            return (
                bg_a * fg_a + self.comp(fg, bg_a) + self.comp(bg, fg_a)
                - min(fg_a * bg_a, ((bg_a * fg_a - bg * fg_a) / fg * bg_a))
            )


class Lighten(BlendMode):
    """
    Lighten blending mode.
    """
    def blend(self, bg: float, fg: float, bg_a: float, fg_a: float) -> float:
        return max(fg * bg_a, bg * fg_a) + self.comp(fg, bg_a) + self.comp(bg, fg_a);
    

class Screen(BlendMode):
    """
    Screen blending mode.
    """
    def blend(self, bg: float, fg: float, bg_a: float, fg_a: float) -> float:
        return 1 - (1 - bg * bg_a) * (1 - fg * fg_a)


class ColorDodge(BlendMode):
    """
    Color Dodge blending mode.
    """
    def blend(self, bg: float, fg: float, bg_a: float, fg_a: float) -> float:
        if fg == fg_a:
            if bg == 0:
                return self.comp(fg, bg_a);
            else:
                return fg_a * bg_a + self.comp(fg, bg_a) + self.comp(bg, fg_a);
        else:
            return min(fg_a * bg_a, bg * (fg_a / (fg_a * bg_a - fg * bg_a)));
        

class Overlay(BlendMode):
    """
    Overlay blending mode.
    """
    def blend(self, bg: float, fg: float, bg_a: float, fg_a: float) -> float:
        if bg * 2 > bg_a:
            return (
                self.comp(fg, bg_a) + self.comp(bg, fg_a) 
                - 2 * (bg_a - bg) * (fg_a - fg) + fg_a * bg_a
            )
        else:
            return fg * bg * 2 + self.comp(fg, bg_a) + self.comp(bg, fg_a)
        

class SoftLight(BlendMode):
    """
    Soft Light blending mode.
    """
    def blend(self, bg: float, fg: float, bg_a: float, fg_a: float) -> float:
        fg_n = fg / fg_a

        if 2 * bg <= bg_a:
            return (
                fg * (bg_a + (2 * bg - bg_a) * (1 - fg_n))
                + self.comp(bg, fg_a) + self.comp(fg, bg_a)
            )
        elif 2 * bg > bg_a and 4 * fg <= fg_a:
            return (
                fg_a * (2 * bg - bg_a) * (16 * fg_n**3 - 12 * fg_n**2 - 3 * fg_n)
                + bg - bg * fg_a + fg
            )
        else:
            return fg_a * (2 * bg - bg_a) * (fg_n**0.5 - fg_n) + bg - bg * fg_a + fg


class HardLight(BlendMode):
    """
    Hard Light blending mode.
    """
    def blend(self, bg: float, fg: float, bg_a: float, fg_a: float) -> float:
        if fg * 2 > fg_a:
            return (
                fg_a * bg_a - 2 * (bg_a - bg) * (fg_a - fg) 
                + self.comp(fg, bg_a) + self.comp(bg, fg_a)
            )
        else:
            return 2 * fg * bg + self.comp(fg, bg_a) + self.comp(bg, fg_a)
        
        
class Difference(BlendMode):
    """
    Difference blending mode.
    """
    def blend(self, bg: float, fg: float, bg_a: float, fg_a: float) -> float:
        return fg + bg - 2 * min(fg * bg_a, bg * fg_a);


class Exclusion(BlendMode):
    """
    Exclusion blending mode.
    """
    def blend(self, bg: float, fg: float, bg_a: float, fg_a: float) -> float:
        return (
            (fg * bg_a + bg * fg_a - 2 * fg * bg) 
            + self.comp(fg, bg_a) + self.comp(bg, fg_a)
        )
    


from typing import Iterable


class Palette:
    """
    `RGBA` Color palette.
    """
    _web: "Palette" = None

    def __init__(self, colors: Iterable) -> None:
        """
        Palette constructor.

        Attributes
        ----------
        colors: `Iterable[RGBA]`
            List of colors.

        Raises
        ------
        `ValueError` if the any of the colors is not `RGBA` instance.
        """
        from .rgba import RGBA

        self._items: list[RGBA] = []
        self._bits: int = 0

        for color in colors:
            self.add(color)

    # magic methods
    def __eq__(self, other) -> bool:
        return isinstance(other, Palette) and all(
            first == second for first, second in zip(self, other))

    def __ne__(self, other) -> bool:
        return not self.__eq__(other)
    
    def __lt__(self, other):
        return isinstance(other, Palette) and len(self.color) < len(other.color)
    
    def __le__(self, other):
        return isinstance(other, Palette) and len(self.color) <= len(other.color)
    
    def __gt__(self, other):
        return not self <= other
    
    def __ge__(self, other):
        return not self < other

    def __str__(self) -> str:
        return f"Palette(num={len(self._items)})"

    def __repr__(self) -> str:
        return f"<Palette colors={self._items}>"

    def __hash__(self) -> int:
        return hash(self.value)
    
    def __getitem__(self, key):
        return self._items[key]
    
    def __iter__(self):
        for item in self._items:
            yield item

    def _check(self, color):
        from .rgba import RGBA
        
        if not isinstance(color, RGBA) or (self._bits and color.bits != self._bits):
            raise ValueError("color must be instance of RGBA and have same bit count")

    # attributes
    def add(self, color) -> None:
        """
        Add a color to the palette.

        Attributes
        ----------
        color: `RGBA`
            Color to add.

        Raises
        ------
        `ValueError` if the color is invalid.
        """
        self._check(color)
        self._items.append(color)

    def remove(self, color) -> None:
        """
        Remove the color to the palette.

        Attributes
        ----------
        color: `RGBA`
            Color to remove.

        Raises
        ------
        `ValueError` if the color is not present or is invalid.
        """
        self._check(color)
        self._items.remove(color)

    # palette generators
    @staticmethod
    def web() -> "Palette":
        """
        Get a palette of web-safe colors.
        """
        from .rgba import RGBA

        if Palette._web is None:
            Palette._web = [
                RGBA((i * 51, j * 51, k * 51)) 
                for i in range(6) 
                for j in range(6) 
                for k in range(6)
            ]
        
        return Palette(Palette._web)
    
    @staticmethod
    def random(num: int) -> "Palette":
        """
        Generate a palette with random colors.

        Attributes
        ----------
        num: `int`
            Number of colors.
        """
        from .hsla import HSLA

        return Palette(HSLA.random().to_rgba() for _ in range(num))
    
    @staticmethod
    def gradient(start, end, num: int) -> "Palette":
        """
        Generate a palette with colors that create gradient from start to end.

        Attributes
        ----------
        start: `RGBA`
            Start color.
        end: `RGBA`
            End color.
        num: `int`
            Number of colors.

        Raises
        ------
        `ValueError` if the number < 2.
        """
        from .rgba import RGBA

        if num < 2:
            raise ValueError("number of colors must be greater than or equal to 2")

        def interpolate(start, end, step):
            return int(start + (end - start) * step)
        
        gradient = []
        for i in range(num):
            step = i / (num - 1) 

            color = RGBA((
                interpolate(start.r, end.r, step),
                interpolate(start.g, end.g, step),
                interpolate(start.b, end.b, step),
                interpolate(start.a, end.a, step)
            ))
           
            gradient.append(color)

        return Palette(gradient)